from __future__ import annotations

import atexit
import json
from typing import cast

import httpx

from ._utils import (
    OpenInferenceAPIInput,
    OpenInferenceAPIOutput,
    OpenInferenceAPIRequestedOutput,
    PydanticOpenInferenceError,
    Singleton,
)


class OpenInferenceHTTPClientAPIError(PydanticOpenInferenceError):
    """Exceptions from the OpenInferenceHTTPClientAPI."""


class BadStatusCodeFromServerError(OpenInferenceHTTPClientAPIError):
    """A bad status code was returned from the open inference server."""

    def __init__(self, status_code: int) -> None:
        self._status_code = status_code
        message = f"Inference server responded with HTTP {status_code}"
        super().__init__(message)

    @property
    def status_code(self) -> int:
        return self._status_code


class OpenInferenceHTTPClientAPI(metaclass=Singleton):
    def __init__(self, base_url: str) -> None:
        if not base_url.endswith("/"):
            base_url += "/"
        self._client = httpx.Client(
            base_url=base_url,
            headers={"Content-Type": "application/json"},
        )
        atexit.register(self._client.close)

    def infer(
        self,
        model_name: str,
        inputs: list[OpenInferenceAPIInput],
        outputs: list[OpenInferenceAPIRequestedOutput] | None = None,
        model_version: str | None = None,
        timeout_seconds: float | None = None,
    ) -> list[OpenInferenceAPIOutput]:
        try:
            payload: dict[str, list[OpenInferenceAPIInput] | list[OpenInferenceAPIRequestedOutput]] = {"inputs": inputs}
            if outputs is not None:
                payload["outputs"] = outputs
            if model_version is None:
                url = f"v2/models/{model_name}/infer"
            else:
                url = f"v2/models/{model_name}/versions/{model_version}/infer"
            response = self._client.post(
                url,
                json=payload,
                timeout=timeout_seconds,
            ).raise_for_status()
            return cast("list[OpenInferenceAPIOutput]", response.json().get("outputs", []))
        except httpx.HTTPStatusError as error:
            raise BadStatusCodeFromServerError(
                status_code=error.response.status_code,
            ) from error
        except (httpx.HTTPError, json.JSONDecodeError) as error:
            raise OpenInferenceHTTPClientAPIError(error) from error



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Generic, TypeVar

if TYPE_CHECKING:
    import sys

    if sys.version_info < (3, 11):
        from typing_extensions import Self
    else:
        from typing import Self

import pydantic

from ._client import OpenInferenceHTTPClientAPI
from ._utils import (
    OpenInferenceAPIInput,
    OpenInferenceAPIOutput,
    OpenInferenceAPIRequestedOutput,
    get_data,
    get_datatype,
    get_shape,
    is_listlike,
    unflatten_data,
)


class InputsBaseModel(pydantic.BaseModel):
    """Base class for model inputs.

    Subclass this class to define the model inputs, e.g.,

        class MyModelInput(InputsBaseModel):
            values: list[int]

    The field names MUST correspond to the names of the model inputs.

    The structure of a field must be mappable to the shape of the
    corresponding input, e.g.,
    - "list[tuple[int, int]]" for a shape of [-1, 2],
    - "tuple[float, float]" for  a shape of [2], or
    - "list[tuple[tuple[int, int], tuple[int, int]]]" for a shape of [-1, 2, 2].
    The actual shape in the input that is sent will have the actual size, not -1.

    NOTE: You can define an input field that is neither a list nor a tuple, just
    a simple datatype like int, float, str; this will correspond to a shape of [1].

    You can use typing.Namedtuple instead of a tuple to get named fields.

    Example:
        class Point(typing.Namedtuple):
            x: float
            y: float

        class ExampleInput(InputsBaseModel):
            points: list[Point]  # name: "points", shape: [-1, 2]
            matrix: tuple[tuple[int, int, int], tuple[int, int, int]]  # name: "matrix", shape: [2, 3]
            text: str  # name: "text", shape: [1]

    The datatype of an input is automatically set depending on the field type, e.g.,
    for a field that is "list[int]" (or "list[tuple[int, int]]" etc) the datatype
    will be "INT64". The actual mapping is

        bool: "BOOL",
        int: "INT64",
        float: "FP32",
        str: "BYTES",

    You can override these defaults using DatatypeOverride, e.g.,

        class ExampleInput(InputsBaseModel):
            points: list[Point]  # datatype: "INT64"
            more_points: Annotated[list[Point], DatatypeOverride("INT32")]  # datatype: "INT32"

    See DatatypeOverride docs for more information.

    NOTE: Because there is just one datatype per input, it means that fields with a mix of types
    like "tuple[int, float, str]" should be avoided.

    """

    model_config = pydantic.ConfigDict(extra="forbid", frozen=True, coerce_numbers_to_str=True)

    def to_inputs(self) -> list[OpenInferenceAPIInput]:
        return [
            {
                "name": field_name,
                "shape": get_shape(field_value),
                "datatype": get_datatype(field_value, type(self).model_fields[field_name]),
                "data": get_data(field_value),
            }
            for field_name, field_value in self.model_dump(mode="json", warnings=False).items()
        ]


class OutputsBaseModel(pydantic.BaseModel):
    """Base class for model outputs.

    Subclass this class to define the model inputs, e.g.,

        class MyModelIOutput(OutputsBaseModel):
            values: list[float]

    The field names MUST correspond to the names of the model outputs.

    The structure of a field must be mappable to the shape of the
    corresponding output, e.g.,
    - "list[tuple[int, int]]" for a shape of [-1, 2],
    - "tuple[float, float]" for  a shape of [2], or
    - "list[tuple[tuple[int, int], tuple[int, int]]]" for a shape of [-1, 2, 2].

    NOTE: You can define an output field that is neither a list nor a tuple, just
    a simple datatype like int, float, str; this will correspond to a shape of [1].

    You can use typing.Namedtuple instead of a tuple to get named fields.

    Example:
        class Point(typing.Namedtuple):
            x: float
            y: float

        class ExampleInput(OutputsBaseModel):
            points: list[Point]  # name: "points", shape: [-1, 2]
            matrix: tuple[tuple[int, int, int], tuple[int, int, int]]  # name: "matrix", shape: [2, 3]
            text: str  # name: "text", shape: [1]

    You can take advantage of pydantic's type coercion by defining the field types
    as you please, including mixed values.

    Example:
        Model outputs of the form
            [
                {
                    "name": "labels",
                    "datatype": "BYTES",
                    "shape": [3, 2],
                    "data": [
                        "SPAM",
                        "0.8",
                        "HAM",
                        "0.3",
                        "SPAM",
                        "0.9"
                    ]
                }
            ]

        can be, using this OutputsBaseModel:

            class LabelAndProbability(typing.Namedtuple):
                label: typing.Literal["SPAM", "HAM"]
                probability: float

            class ExampleInput(OutputsBaseModel):
                labels: list[LabelAndProbability]

        coerced to

            ExampleInput(
                labels=[
                    LabelAndProbability(label="SPAM", probability=0.8),
                    LabelAndProbability(label="HAM", probability=0.3),
                    LabelAndProbability(label="SPAM", probability=0.9),
                ]
            )

        i.e., we get named fields of the desired type.

    """

    model_config = pydantic.ConfigDict(extra="forbid", frozen=True, coerce_numbers_to_str=True)

    @classmethod
    def from_outputs(cls: type[Self], outputs: list[OpenInferenceAPIOutput]) -> Self:
        values: dict[str, Any] = {}
        for output in outputs:
            value = unflatten_data(output["shape"], output["data"])
            if output["shape"] == [1] and not is_listlike(cls.model_fields[output["name"]].annotation):
                values[output["name"]] = value[0]
            else:
                values[output["name"]] = value
        return cls(**values)

    @classmethod
    def get_requested_outputs(cls) -> list[OpenInferenceAPIRequestedOutput]:
        return [{"name": field_name} for field_name in cls.model_fields]


InputsModelT = TypeVar("InputsModelT", bound=InputsBaseModel)
OutputsModelT = TypeVar("OutputsModelT", bound=OutputsBaseModel)


class RemoteModel(Generic[InputsModelT, OutputsModelT]):
    """Interface toward a remote model called via the Open Inference protocol.

    Define the server, model, inputs, and outputs, then call
    the "infer" method to call the remote model, automatically
    handling conversion between the inputs/outputs models and
    the Open Inference protocol.

    This class is designed to be thread-safe.

    """

    def __init__(  # noqa: PLR0913
        self,
        *,
        model_name: str,
        inputs_model: type[InputsModelT],
        outputs_model: type[OutputsModelT],
        server_url: str,
        request_timeout_seconds: float | None = None,
        model_version: str | None = None,
    ) -> None:
        """Initialize a RemoteModel.

        Args:
            model_name: The name of the model in the inference server.
            inputs_model: InputsBaseModel subclass defining the model inputs.
            outputs_model: OutputsBaseModel subclass defining the model outputs.
            server_url: The URL of the inference server (not including the
                API version "v2/..." and onward.)
            request_timeout_seconds: Optional timeout (seconds) for requests to the inference server.
            model_version: Optional model version.

        """
        self._model_name = model_name
        self._model_version = model_version
        self._inputs_model = inputs_model
        self._outputs_model = outputs_model
        self._client_API = OpenInferenceHTTPClientAPI(base_url=server_url)
        self._request_timeout_seconds = request_timeout_seconds

    def infer(self, inputs: InputsModelT) -> OutputsModelT:
        """Use the remote model for inference.

        Args:
            inputs: The model inputs in the form of the InputsBaseModel
                subclass defined for this RemoteModel instance.

        Returns:
            outputs: The model outputs in the form of the OutputsBaseModel
                subclass defined for this RemoteModel instance.

        """
        if not isinstance(inputs, self._inputs_model):
            raise TypeError(f"Bad inputs type: {type(inputs)}")  # noqa: TRY003
        return self._outputs_model.from_outputs(
            self._client_API.infer(
                model_name=self._model_name,
                model_version=self._model_version,
                inputs=inputs.to_inputs(),
                outputs=self._outputs_model.get_requested_outputs(),
                timeout_seconds=self._request_timeout_seconds,
            )
        )



from __future__ import annotations

import itertools
import sys
import types
from typing import TYPE_CHECKING, Any, ClassVar, Final, Literal, TypedDict

if sys.version_info < (3, 11):
    from typing_extensions import NotRequired  # pragma: no cover
else:
    from typing import NotRequired  # pragma: no cover

if sys.version_info < (3, 12):
    from typing_extensions import override  # pragma: no cover
else:
    from typing import override  # pragma: no cover

if sys.version_info < (3, 13):
    from more_itertools import batched  # pragma: no cover
else:
    from itertools import batched  # pragma: no cover
if TYPE_CHECKING:
    from collections.abc import Mapping

    from pydantic.fields import FieldInfo

Shape = list[int]
Datatype = Literal[
    "BOOL",
    "UINT8",
    "UINT16",
    "UINT32",
    "UINT64",
    "INT8",
    "INT16",
    "INT32",
    "INT64",
    "FP16",
    "FP32",
    "FP64",
    "BYTES",
]
Data = list[Any]
_TYPE_TO_DATATYPE_MAP: Final[Mapping[type, Datatype]] = {
    bool: "BOOL",
    int: "INT64",
    float: "FP32",
    str: "BYTES",
}


class PydanticOpenInferenceError(Exception):
    """Package base exception."""


class ShapeDataMismatchError(PydanticOpenInferenceError):
    """Exception raised when shape and data do not match."""


class DatatypeOverride:
    """Use with typing.Annotated to override the default datatype in inputs.

    In this example InputsBaseModel, the datatype of
    "values" would be "INT64":

        class IntInputsBaseModel(InputsBaseModel):
            values: list[int]

    Using DatatypeOverride, we can instead force the datatype
    to be anything we want, e.g., "INT16":

        class IntInputsBaseModel(InputsBaseModel):
            values: Annotated[list[int], DatatypeOverride("INT16")]

    Note that this simply sets the datatype as given. There
    are no additional checks (for sign, size, etc).

    """

    __slots__ = ("_datatype",)

    def __init__(self, datatype: Datatype) -> None:
        self._datatype = datatype

    @property
    def datatype(self) -> Datatype:
        return self._datatype


def is_flat(data: Data) -> bool:
    return not data or not any(isinstance(x, (list, tuple)) for x in data)


def parse_row_major_order(shape: Shape, data: list[Any]) -> list[Any]:
    if len(shape) == 1:
        if shape[0] != len(data):
            raise ShapeDataMismatchError
        return data
    new_shape: Shape = list(shape[:-1])
    return parse_row_major_order(new_shape, list(batched(data, n=shape[-1], strict=True)))


def is_listlike(annotation: type[Any] | None) -> bool:
    if annotation is None:
        return False
    if isinstance(annotation, types.GenericAlias):
        annotation = annotation.__origin__  # type: ignore[unreachable]
    return any(t in (list, tuple) for t in itertools.chain((annotation,), annotation.__bases__))


def unflatten_data(shape: Shape, data: Data) -> Data:
    if is_flat(data):
        return parse_row_major_order(shape, data)
    return data


def get_shape(value: Any) -> Shape:
    shape: Shape = []
    while isinstance(value, (list, tuple)):
        shape.append(len(value))
        value = value[0]
    return shape or [1]


def get_datatype(value: Any, field_info: FieldInfo) -> Datatype:
    overrides: list[DatatypeOverride] = [x for x in field_info.metadata if isinstance(x, DatatypeOverride)]
    if overrides:
        return overrides[0].datatype
    while isinstance(value, (list, tuple)):
        value = value[0]
    return _TYPE_TO_DATATYPE_MAP[type(value)]


def get_data(value: Any) -> Data:
    if not isinstance(value, (tuple, list)):
        return [value]
    return list(value)


class _OpenInferenceAPIPut(TypedDict):
    name: str
    shape: Shape
    datatype: Datatype
    data: Data


class OpenInferenceAPIInput(_OpenInferenceAPIPut):
    parameters: NotRequired[dict[str, Any]]


class OpenInferenceAPIRequestedOutput(TypedDict):
    name: str
    parameters: NotRequired[dict[str, Any]]


class OpenInferenceAPIOutput(_OpenInferenceAPIPut):
    parameters: NotRequired[dict[str, Any]]


class Singleton(type):
    _instances: ClassVar[dict[tuple[type, tuple[Any, ...], str], Singleton]] = {}

    @override
    def __call__(cls, *args, **kwargs):  # type: ignore[no-untyped-def]
        key = (cls, args, str(kwargs))
        if key not in cls._instances:
            cls._instances[key] = super().__call__(*args, **kwargs)
        return cls._instances[key]



"""Package for wrapping calls to an Open Inference server REST API."""

from ._client import BadStatusCodeFromServerError
from ._remote_model import InputsBaseModel, OutputsBaseModel, RemoteModel
from ._utils import DatatypeOverride, PydanticOpenInferenceError

__all__ = (
    "BadStatusCodeFromServerError",
    "DatatypeOverride",
    "InputsBaseModel",
    "OutputsBaseModel",
    "PydanticOpenInferenceError",
    "RemoteModel",
)
